package de.unisaar.jfap.ex9;

import de.unisaar.jfap.ex7.DirectedGraph;
import de.unisaar.jfap.ex7.Edge;
import de.unisaar.jfap.ex8.DijkstraShortestPath;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.List;
import java.util.Map;

/**
 * Find shortest path in elevated terrain.
 *
 * @author Jörg Steffen, DFKI
 * @author Bernd Kiefer, DFKI
 */
public final class PathFinder {

  // graph with elevation as vertex property and constant edge weight of 1
  private static DirectedGraph<Integer, Integer> graph;
  // start and end vertex of path
  private static int start;
  private static int end;

  private PathFinder() {
    // private constructor to enforce noninstantiability
  }

  /**
   * Create edge between source vertex and target vertex if target vertex is max one elevation
   * higher than source vertex (forward) or vice versa (backward).
   *
   * @param source source vertex
   * @param target target vertex which is a direct neighbor of source vertex
   * @param forward flag indicating if we consider path from start to end (forward) or end to start
   *     (backward)
   */
  private static void createEdge(int source, int target, boolean forward) {
    int sourceElevation = graph.getVertexProperty(source);
    int targetElevation = graph.getVertexProperty(target);
    if (sourceElevation >= targetElevation - 1) {
      // create edge with weight 1
      if (forward) {
        graph.newEdge(1, source, target);
      } else {
        graph.newEdge(1, target, source);
      }
    }
  }

  /**
   * Read map from given path and initialize graph with start and end vertex as well as elevation
   * map.
   *
   * @param path path where to read the map from
   * @param forward flag indicating if we consider path from start to end (forward) or end to start
   *     (backward)
   * @throws IOException if map reading fails
   */
  private static void init(Path path, boolean forward) throws IOException {

    graph = new DirectedGraph<>();
    List<String> lines = Files.readAllLines(path);
    // mapping of original map to graph vertices (integer)
    int[][] map = new int[lines.size()][lines.get(0).length()];

    for (int i = 0; i < lines.size(); i++) {
      String oneLine = lines.get(i);
      for (int j = 0; j < oneLine.length(); j++) {
        char elevation = oneLine.charAt(j);
        // we normalize elevation so that a = 0, b = 1, ..., z = 25
        switch (elevation) {
          case 'S':
            start = graph.newVertex(0);
            map[i][j] = start;
            break;
          case 'E':
            end = graph.newVertex('z' - 'a');
            map[i][j] = end;
            break;
          default:
            map[i][j] = graph.newVertex(elevation - 'a');
        }
      }
    }

    // create edges if elevation difference between source and target vertex is in allowed range
    for (int i = 0; i < map.length; i++) {
      for (int j = 0; j < map[i].length; j++) {
        int source = map[i][j];
        if (i > 0) {
          // check edge to north
          createEdge(source, map[i - 1][j], forward);
        }
        if (i < map.length - 1) {
          // check edge to south
          createEdge(source, map[i + 1][j], forward);
        }
        if (j > 0) {
          // check edge to east
          createEdge(source, map[i][j - 1], forward);
        }
        if (j < map[i].length - 1) {
          // check edge to west
          createEdge(source, map[i][j + 1], forward);
        }
      }
    }
  }

  /**
   * Read map from given path and return the length of the shortest path from start(S) to end(E)
   * location on the map.
   *
   * @param path path where to read the map from
   * @return length of shortest path
   * @throws IOException if map reading fails
   */
  public static int findShortestPathLength(Path path) throws IOException {
    init(path, true);
    DijkstraShortestPath dsp = new DijkstraShortestPath(graph);
    List<Edge<Integer>> shortestPath = dsp.shortestPath(start, end);
    return shortestPath.size();
  }

  /**
   * Read map from given path and return the length of the shortest path from start(S) to any of the
   * 'a' locations on the map.
   *
   * @param path path where to read the map from
   * @return length of shortest path
   * @throws IOException if map reading fails
   */
  public static int findShortestPathLengthForAs(Path path) throws IOException {
    init(path, false);
    DijkstraShortestPath dsp = new DijkstraShortestPath(graph);
    // we use an invalid target vertex to force Dijkstra to calculate the shortest paths
    // from the *end* vertex to *all* other vertices
    dsp.shortestPath(end, -1);
    // find to shortest path from end to an 'a' location (having an elevation of 0)
    int minPathLength = Integer.MAX_VALUE;
    Map<Integer, Integer> distances = dsp.getDistances();
    for (var oneEntry : distances.entrySet()) {
      int vertex = oneEntry.getKey();
      if (graph.getVertexProperty(vertex) == 0) {
        minPathLength = Math.min(minPathLength, oneEntry.getValue());
      }
    }
    return minPathLength;
  }
}
